import unittest
from src.masonite.headers import HeaderBag, Header


class TestHeaders(unittest.TestCase):
    def test_headers_can_be_registered(self):
        bag = HeaderBag()
        bag.add(Header("Content-Type", "application/json"))
        self.assertTrue(bag["CONTENT_TYPE"])

    def test_headers_can_be_rendered(self):
        bag = HeaderBag()
        bag.add(Header("content-type", "application/json"))
        bag.add(Header("x-forwarded-for", "127.0.0.1"))
        self.assertEqual(
            bag.render(),
            [("Content-Type", "application/json"), ("X-Forwarded-For", "127.0.0.1")],
        )

    def test_headers_can_check_with_in(self):
        bag = HeaderBag()
        bag.add(Header("Content-Type", "application/json"))

        self.assertTrue("Content-Type" in bag)

    def test_headers_can_be_retrieved(self):
        bag = HeaderBag()
        bag.add(Header("Content-Type", "application/json"))
        self.assertEqual(bag.get("content-type").value, "application/json")
        self.assertEqual(bag.get("Content-Type").value, "application/json")

    def test_convert_name_back(self):
        bag = HeaderBag()
        self.assertEqual(bag.convert_name_back("X-Rate-Limited"), "HTTP_X_RATE_LIMITED")
        self.assertEqual(bag.convert_name_back("content-type"), "CONTENT_TYPE")
        self.assertEqual(bag.convert_name_back("Remote-Addr"), "REMOTE_ADDR")

    def test_headers_can_be_output_as_dict(self):
        bag = HeaderBag()
        bag.add(Header("content-type", "application/json"))
        bag.add(Header("x-forwarded-for", "127.0.0.1"))
        self.assertEqual(
            bag.to_dict(),
            {
                "Content-Type": "application/json",
                "X-Forwarded-For": "127.0.0.1",
            },
        )

        self.assertEqual(
            bag.to_dict(server_names=True),
            {
                "CONTENT_TYPE": "application/json",
                "HTTP_X_FORWARDED_FOR": "127.0.0.1",
            },
        )



from tests import TestCase

from src.masonite.rates import GlobalLimiter, GuestsOnlyLimiter, Limiter, Limit
from src.masonite.routes import Route
from src.masonite.facades import RateLimiter
from src.masonite.exceptions import ThrottleRequestsException
from tests.integrations.app.User import User


class CustomLimit(Limiter):
    def allow(self, request):
        return Limit.per_day(2)

    def get_response(self, request, response, headers):
        return response.view("Too many attempts, please try again tomorrow", 429)


class TestLimitingRequest(TestCase):
    def setUp(self):
        super().setUp()
        # register limiters
        RateLimiter.register("test_global", GlobalLimiter("4/minute"))
        RateLimiter.register("test_guests", GuestsOnlyLimiter("2/minute"))
        RateLimiter.register("test_custom", CustomLimit())
        self.addRoutes(
            Route.get("/throttled/global", "WelcomeController@show").middleware(
                "throttle:test_global"
            ),
            Route.get("/throttled/guests", "WelcomeController@show").middleware(
                "throttle:test_guests"
            ),
            Route.get("/throttled/arg", "WelcomeController@show").middleware(
                "throttle:2/hour"
            ),
            Route.get("/throttled/custom", "WelcomeController@show").middleware(
                "throttle:test_custom"
            ),
        )

    def tearDown(self):
        super().tearDown()
        # delete cache entries between tests for idempotent tests
        self.application.make("cache").store().flush()
        # always restore time if it has been faked
        self.restoreTime()

    def test_limit_from_middleware_arg(self):
        self.get("/throttled/arg").assertOk()
        self.get("/throttled/arg").assertOk()
        with self.assertRaises(ThrottleRequestsException):
            self.get("/throttled/arg")
        # set time 2 hours later
        self.fakeTimeInFuture(2, "hours")
        self.get("/throttled/arg").assertOk()

    def test_limiter_in_production(self):
        self.withExceptionsHandling()
        with self.debugMode(False):
            count = 0
            for _ in range(4):
                count += 1
                (
                    self.get("/throttled/global")
                    .assertOk()
                    .assertHasHeader("X-RateLimit-Limit", "4")
                    .assertHasHeader("X-RateLimit-Remaining", str(4 - count))
                )

            (
                self.get("/throttled/global")
                .assertLimited()
                .assertHasHeader("Retry-After")
                .assertHasHeader("X-RateLimit-Reset")
                .assertHasHeader("X-RateLimit-Limit", "4")
                .assertHasHeader("X-RateLimit-Remaining", "0")
            )

    def test_limit_guests_users_only(self):
        # can make only 2 requests in one minute
        self.get("/throttled/guests").assertOk().assertHasHeader(
            "X-RateLimit-Limit", "2"
        )
        self.get("/throttled/guests").assertOk()
        with self.assertRaises(ThrottleRequestsException):
            self.get("/throttled/guests")
        # no limits if authenticated, wait for PR#551 to be merged which fix actingAs
        # self.actingAs(User.first()).get(
        #     "/throttled/guests"
        # ).assertOk().assertHeaderMissing("X-RateLimit-Limit")

    def test_using_custom_response(self):
        self.get("/throttled/custom").assertOk()
        self.get("/throttled/custom").assertOk()
        self.get("/throttled/custom").assertLimited().assertContains(
            "Too many attempts, please try again tomorrow"
        )



from typing import Any, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from ..rates.limiters import Limiter

class RateLimiter:
    """Rate Limiter facades to add rate limiting to your functions."""

    def register(self, name, callback: "Limiter") -> "RateLimiter":
        """Register a new rate limiter with the given name"""
        ...
    def attempts(self, key: str) -> int:
        """Get number of attempts left for a given rate limiter key."""
        ...
    def get_limiter(self, name: str) -> "Limiter":
        """Get rate limiter registered with the given name."""
        ...
    def attempt(
        key: str, callback: Callable, max_attempts: int, delay: int = 60
    ) -> Any:
        """Try to execute the given callback if not limited by the 'key' rate limiter."""
        ...
    def too_many_attempts(self, key: str, max_attempts: int) -> bool:
        """Check if given rate limiter key got more (or equal) attempts than max_attempts."""
        ...
    def hit(self, key: str, delay: int) -> int:
        """Add one attempt for the given key."""
        ...
    def reset_attempts(self, key: str) -> bool:
        """Reset attempts count to 0 for the given key."""
        ...
    def clear(self, key: str):
        """Clear all data of the given rate limiter key."""
        ...
    def available_at(self, key: str) -> int:
        """Get UNIX integer timestamp at which rate limiter key will be available again."""
        ...
    def available_in(self, key: str) -> int:
        """Get seconds in which rate limiter key will be available again."""
        ...
    def remaining(self, key: str, max_attempts: int) -> int:
        """Get remaining attempts before given rate limiter key is limited regarding max_attempts limit."""
        ...



import pendulum
from tests import TestCase

from src.masonite.facades import RateLimiter, Cache


def my_function():
    return "done"


class TestRateLimiter(TestCase):
    def tearDown(self):
        super().tearDown()
        # delete cache entries between tests for idempotent tests
        self.application.make("cache").store().flush()
        # always restore time if it has been faked
        self.restoreTime()

    def test_attempt(self):
        self.assertEqual(RateLimiter.attempt("test_attempt", my_function, 2), "done")
        self.assertEqual(RateLimiter.attempt("test_attempt", my_function, 2), "done")
        self.assertEqual(RateLimiter.attempt("test_attempt", my_function, 2), False)

    def test_clear_remove_cache_keys(self):
        RateLimiter.attempt("test_key", my_function, 2)
        assert Cache.store().has("test_key")
        assert Cache.store().has("test_key-timer")
        RateLimiter.clear("test_key")
        assert not Cache.store().has("test_key-timer")
        assert not Cache.store().has("test_key")

    def test_reset_attempts(self):
        RateLimiter.attempt("test_key", my_function, 2)
        assert Cache.store().get("test_key") == "1"
        RateLimiter.reset_attempts("test_key")
        assert Cache.store().get("test_key") == "0"

    def test_attempts(self):
        RateLimiter.attempt("test_key", my_function, 2)
        RateLimiter.attempt("test_key", my_function, 2)
        assert RateLimiter.attempts("test_key") == 2

    def test_remaining(self):
        RateLimiter.attempt("test_key", my_function, 2)
        assert RateLimiter.remaining("test_key", 2) == 1

    def test_hit(self):
        now = pendulum.now()
        self.fakeTime(now)
        should_be_available_at = now.add(seconds=40)
        RateLimiter.hit("test_key", 40)
        assert Cache.store().get("test_key") == "1"
        assert Cache.store().get("test_key-timer") == str(
            should_be_available_at.int_timestamp
        )

    def test_available_at(self):
        now = pendulum.now()
        self.fakeTime(now)
        should_be_available_at = now.add(seconds=40)
        RateLimiter.hit("test_key", 40)
        assert (
            RateLimiter.available_at("test_key") == should_be_available_at.int_timestamp
        )



from tests import TestCase
import time
import pytest


@pytest.mark.integrations
class TestRedisSession(TestCase):
    def setUp(self):
        super().setUp()
        self.session = self.application.make("session")
        self.driver = self.session.driver("redis")
        self.session.start("redis")

    def teardown(self):
        self.session.flush()
        self.session.close()
        self.session = None

    def test_can_increment(self):
        self.session.set("count", "1")
        self.assertEqual(self.session.get("count"), "1")
        self.session.increment("count")
        self.assertEqual(self.session.get("count"), "2")
        self.session.decrement("count")
        self.assertEqual(self.session.get("count"), "1")

    def test_delete(self):
        self.session.set("forget", "1")
        self.assertEqual(self.session.get("forget"), "1")
        self.session.delete("forget")
        self.assertEqual(self.session.get("forget"), None)

    def test_set(self):
        self.session.set("remember", "1")
        self.assertEqual(self.session.get("remember"), "1")

    def test_can_flash(self):
        self.session.flash("key", "test")
        self.assertEqual(self.session.get("key"), "test")
        self.assertEqual(self.session.get("key"), None)

    def test_can_pull_session(self):
        self.session.set("key", "test")
        self.assertEqual(self.session.get("key"), "test")

        key = self.session.pull("key")
        self.assertEqual(key, "test")
        self.assertEqual(self.session.get("key"), None)

    def test_flush(self):
        self.session.set("dic", {"id": 1, "name": "Joe"})
        self.session.flush()
        self.assertIsNone(self.session.get("dic"))



import json
from typing import Any, TYPE_CHECKING

import pendulum as pdlm
if TYPE_CHECKING:
    from redis import Redis


class RedisDriver:
    def __init__(self, application):
        self.application = application
        self.connection = None
        self.options = {}
        self._internal_cache: dict = None

    def set_options(self, options: dict) -> "RedisDriver":
        self.options = options
        return self

    def get_connection(self) -> "Redis":
        if self.connection:
            return self.connection

        try:
            from redis import Redis
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'redis' library. Run 'pip install redis' to fix this."
            )

        self.connection = Redis(
            **self.options.get("options", {}),
            host=self.options.get("host"),
            port=self.options.get("port"),
            password=self.options.get("password"),
            decode_responses=True,
        )

        return self.connection

    def _load_from_store(self) -> None:
        """
        copy all the "cache" key value pairs for faster access
        """
        if self._internal_cache is not None:
            return

        self._internal_cache = {}

        cursor = "0"
        prefix = self.get_cache_namespace()
        while cursor != 0:
            cursor, keys = self.get_connection().scan(
                cursor=cursor, match=prefix + "*", count=100000
            )
            if keys:
                values = self.get_connection().mget(*keys)
                store_data = dict(zip(keys, values))
                for key, value in store_data.items():
                    key = key.replace(prefix, "")
                    value = self.unpack_value(value)
                    # we dont load the ttl (expiry)
                    # because there is an O(N) performance hit
                    self._internal_cache[key] = {
                        "value": value,
                        "expires": None,
                    }

    def get_cache_namespace(self) -> str:
        """
        Build the prefix for the key to use in Redis
        """
        namespace = self.options.get("name", None)
        namespace += ":" if namespace else ""
        return f"{namespace}cache:"

    def add(self, key: str, value: Any = None) -> Any:
        if value is None:
            return None

        self.put(key, value)
        return value

    def get(self, key: str, default: Any = None, **options) -> Any:
        self._load_from_store()
        if not self.has(key):
            return default or None

        key_expiry = self._internal_cache[key].get("expires", None)
        if key_expiry is None:
            # the ttl value can also provide info on the
            # existence of the key in the store
            ttl = self.get_connection().ttl(key)
            if ttl == -1:
                # key exists but has no set ttl
                ttl = self.get_default_timeout()
            elif ttl == -2:
                # key not found in store
                self._internal_cache.pop(key)
                return default or None

            key_expiry = self._expires_from_ttl(ttl)
            self._internal_cache[key]["expires"] = key_expiry

        if pdlm.now() > key_expiry:
            # the key has expired so remove it from the cache
            self._internal_cache.pop(key)
            return default or None

        # the key has not yet expired
        return self._internal_cache.get(key)["value"]

    def put(self, key: str, value: Any = None, seconds: int = None, **options) -> Any:
        if not key or value is None:
            return None

        store_value = value
        if isinstance(value, (dict, list, tuple)):
            store_value = json.dumps(value)
        elif isinstance(value, int):
            store_value = str(value)

        self._load_from_store()
        key_ttl = seconds or self.get_default_timeout()
        self.get_connection().set(
            f"{self.get_cache_namespace()}{key}", store_value, ex=key_ttl
        )
        expires = self._expires_from_ttl(key_ttl)
        self._internal_cache.update({
            key: {
                "value": value,
                "expires": expires,
            }
        })

    def has(self, key: str) -> bool:
        self._load_from_store()
        return key in self._internal_cache

    def increment(self, key: str, amount: int = 1) -> int:
        value = int(self.get(key)) + amount
        self.put(key, value)
        return value

    def decrement(self, key: str, amount: int = 1) -> int:
        value = int(self.get(key)) - amount
        self.put(key, value)
        return value

    def remember(self, key: str, callable):
        value = self.get(key)

        if value:
            return value

        callable(self)

        return self.get(key)

    def forget(self, key: str) -> None:
        if not self.has(key):
            return
        self.get_connection().delete(f"{self.get_cache_namespace()}{key}")
        self._internal_cache.pop(key)

    def flush(self) -> None:
        self.get_connection().flushall()
        self._internal_cache = None

    def get_default_timeout(self) -> int:
        # if unset default timeout of cache vars is 1 month
        return int(self.options.get("timeout", 60 * 60 * 24 * 30))

    def unpack_value(self, value: Any) -> Any:
        value = str(value)
        if value.isdigit():
            return int(value)

        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            return value

    def _expires_from_ttl(self, ttl: int) -> pdlm.DateTime:
        return pdlm.now().add(seconds=ttl)



from tests import TestCase

from src.masonite.response import Response


class TestResponse(TestCase):
    def test_with_headers(self):
        response = Response(self.application)
        response = response.with_headers({"X-Test": "value", "X-Rate-Limited": "true"})
        self.assertIsInstance(response, Response)
        self.assertEqual(response.header("X-Test"), "value")
        self.assertEqual(response.header("X-Rate-Limited"), "true")

    def test_encoding_chinese_characters(self):
        response = Response(self.application)
        content = response.json({"test": "我"})
        self.assertEqual(content.decode("utf-8"), '{"test": "我"}')

    def test_redirect_external(self):
        response = Response(self.application)
        response = response.redirect("https://google.com")
        self.assertEqual(response._status, "302 Found")
        self.assertEqual(response.header_bag.get("Location").value, "https://google.com")



from src.masonite.middleware import SessionMiddleware
from src.masonite.routes import Route
from tests import TestCase


class TestResponseHelpers(TestCase):
    def setUp(self):
        super().setUp()
        self.addRoutes(
            Route.get("/test-with-errors", "WelcomeController@with_errors"),
            Route.get("/test-with-input", "WelcomeController@form_with_input"),
        )

    def test_with_input(self):
        self.get("/test-with-input", {"name": "Sam"}).assertSessionHas("name", "Sam")

    def test_with_errors(self):
        self.get("/test-with-errors").assertSessionHasErrors().assertSessionHasErrors(
            ["email"]
        )


class TestResponseWithMiddleware(TestCase):
    def setUp(self):
        super().setUp()
        self.application.make("middleware").add(
            {
                "web": SessionMiddleware,
            }
        )
        self.addRoutes(
            Route.get("/test-with-errors", "WelcomeController@with_errors").middleware(
                "web"
            ),
            Route.get(
                "/test-with-input", "WelcomeController@form_with_input"
            ).middleware("web"),
        )

    def test_with_input(self):
        (
            self.get("/test-with-input", {"name": "Sam"})
            .assertSessionHas("name", "Sam")
            .assertSessionHasNoErrors()
        )

    def test_with_errors(self):
        (
            self.get("/test-with-errors")
            .assertSessionHasErrors()
            .assertSessionHasErrors(["email"])
            .assertSessionHasNoErrors()
        )



from tests import TestCase
from src.masonite.routes import Route
from src.masonite.middleware import CorsMiddleware


class TestCors(TestCase):
    def setUp(self):
        super().setUp()
        # add CorsMiddleware
        self.application.make("middleware").add([CorsMiddleware])
        self.cors = self.application.make("cors")
        self.default_options = {
            "allowed_origins": ["http://localhost"],
            "allowed_headers": ["X-Test-1", "X-Test-2"],
            "allowed_methods": ["GET", "POST"],
            "exposed_headers": [],
            "max_age": 0,
            "support_credentials": False,
            "paths": ["api/*"],
        }
        self.cors.set_options(self.default_options)
        self.addRoutes(
            Route.post(
                "/api/test", "tests.integrations.controllers.api.TestController@show"
            ),
            Route.put(
                "/api/test", "tests.integrations.controllers.api.TestController@show"
            ),
            Route.get("/", "WelcomeController@test"),
            Route.post("/", "WelcomeController@api"),
        )

    def tearDown(self):
        super().tearDown()
        # reset CORS options to test default
        self.cors.set_options(self.default_options)

    def test_response_when_origin_not_allowed(self):
        self.withHeaders(
            {
                "Origin": "http://djangoproject.com",
                "Access-Control-Request-Method": "POST",
            }
        ).options("/api/test").assertNoContent().assertHasHeader(
            "Access-Control-Allow-Origin", "http://localhost"
        )

    def test_access_control_allow_origin_if_no_origin(self):
        self.withHeaders({"Access-Control-Request-Method": "POST",}).options(
            "/api/test"
        ).assertNoContent().assertHasHeader(
            "Access-Control-Allow-Origin", "http://localhost"
        )

    def test_access_control_allow_origin_if_origin(self):
        self.withHeaders(
            {
                "Origin": "http://localhost",
                "Access-Control-Request-Method": "POST",
            }
        ).options("/api/test").assertNoContent().assertHasHeader(
            "Access-Control-Allow-Origin", "http://localhost"
        )

    def test_allow_all_origins(self):
        self.cors.set_options({**self.default_options, "allowed_origins": ["*"]})

        self.withHeaders(
            {
                "Origin": "https://masoniteproject.com",
                "Access-Control-Request-Method": "POST",
            }
        ).options("/api/test").assertNoContent().assertHasHeader(
            "Access-Control-Allow-Origin", "*"
        )

    def test_response_when_method_allowed(self):
        self.withHeaders(
            {
                "Origin": "http://localhost",
                "Access-Control-Request-Method": "POST",
            }
        ).post("/api/test").assertOk().assertContains(
            "welcome"
        ).dumpResponseHeaders().assertHeaderMissing(
            "Access-Control-Allow-Methods"
        )

    def test_response_when_method_not_allowed(self):
        self.withHeaders(
            {
                "Origin": "http://localhost",
                "Access-Control-Request-Method": "PUT",
            }
        ).post("/api/test").assertOk().assertHeaderMissing(
            "Access-Control-Allow-Methods"
        )

    def test_allow_headers(self):
        self.withHeaders(
            {
                "Origin": "https://masoniteproject.com",
                "Access-Control-Request-Method": "POST",
                "Access-Control-Request-Headers": "x-test-1, x-test-2",
            }
        ).options("/api/test").assertNoContent().assertHasHeader(
            "Access-Control-Allow-Headers", "x-test-1, x-test-2"
        )

    def test_allow_headers_when_header_not_allowed(self):
        self.withHeaders(
            {
                "Origin": "https://masoniteproject.com",
                "Access-Control-Request-Method": "POST",
                "Access-Control-Request-Headers": "x-other-3",
            }
        ).options("/api/test").assertNoContent().assertHasHeader(
            "Access-Control-Allow-Headers", "x-test-1, x-test-2"
        )

    def test_response_with_allowed_header(self):
        self.withHeaders(
            {
                "Origin": "http://localhost",
                "Access-Control-Request-Headers": "X-Test-1",
            }
        ).post("/api/test").assertOk().assertContains("welcome").assertHeaderMissing(
            "Access-Control-Allow-Headers"
        )

    def test_allow_all_origins_with_wildcard(self):
        self.cors.set_options(
            {**self.default_options, "allowed_origins": ["*.masoniteproject.com"]}
        )

        self.withHeaders(
            {
                "Origin": "https://docs.masoniteproject.com",
                "Access-Control-Request-Method": "POST",
            }
        ).options("/api/test").assertNoContent().dumpResponseHeaders().assertHasHeader(
            "Access-Control-Allow-Origin", "*.masoniteproject.com"
        )



from .SayHi import SayHello
from .Repository import Repository


class GreetingService:
    def __init__(self, say_hello: SayHello):
        self.say_hello = say_hello

    def handle(self, repository: Repository):
        repository.user.name = "Jack Sparrow"
        return repository.user



from tests import TestCase
from src.masonite.validation import email, isnt, is_in, numeric
from src.masonite.validation import Validator


class TestValidation(TestCase):
    def test_can_validate_request(self):
        request = self.make_request(query_string="email=joe@masoniteproject.com")
        validation = request.validate(
            {
                "email": "required",
            }
        )

        self.assertEqual(validation.all(), {})

    def test_can_validate_request_with_no_inputs(self):
        request = self.make_request(query_string="")
        validation = request.validate(
            {
                "email": "required",
            }
        )

        self.assertEqual(validation.all(), {"email": ["The email field is required."]})

    def test_custom_messages(self):
        request = self.make_request(post_data={"not_email": "joe@masonite.com", "age": 20})
        # validator = Validator()
        validate = request.validate(
            isnt(is_in("age", [20, 21]), numeric("age")),
            isnt(email("not_email")),
            {
                "username": "required",
                "valid_email": "required|email",
                "secondary_email": "required|email",
            },
            messages={
                "username.required": "Custom Message for required Username.",
                "not_email.isnt_email": "Custom Message for Email not being an email.",
                "valid_email.required": "Custom Message for required Email.",
                "valid_email.email": "Custom Message for Email being a valid email.",
                "age.isnt_is_in.isnt_numeric": "Custom: Age must not be in 20, 21 and must not be numeric.",
            },
        )

        self.assertIn(
            "Custom: Age must not be in 20, 21 and must not be numeric.",
            validate.get("age"),
        )
        self.assertIn(
            "Custom Message for required Email.",
            validate.get("valid_email"),
        )
        self.assertIn(
            "Custom Message for Email being a valid email.",
            validate.get("valid_email"),
        )
        self.assertIn(
            "Custom Message for Email not being an email.",
            validate.get("not_email"),
        )
        self.assertIn(
            "Custom Message for required Username.",
            validate.get("username"),
        )
        self.assertIn(
            "The secondary_email field is required.",
            validate.get("secondary_email"),
        )
        self.assertIn(
            "The secondary_email must be a valid email address.",
            validate.get("secondary_email"),
        )

    def test_can_forward_validation_calls(self):
        request = self.make_request(query_string="")
        validate = Validator()

        errors = request.validate(
            validate.required(['user', 'email']),
            validate.accepted('terms')
        )

        self.assertIn("The user field is required.", errors.get("user"))
        self.assertIn("The email field is required.", errors.get("email"))
        self.assertIn("The terms must be accepted.", errors.get("terms"))




"""Security Related Config"""

# Cross-Origin Resource Sharing
CORS = {
    "paths": ["api/*"],
    "allowed_methods": ["*"],
    "allowed_origins": ["*"],
    "allowed_headers": ["*"],
    "exposed_headers": [],
    "max_age": None,
    "supports_credentials": False,
}



import os


DRIVERS = {
    "default": "cookie",
    "cookie": {},
    "redis": {
        "host": "127.0.0.1",
        "port": 6379,
        "password": "",
        "options": {"db": 1},  # redis module driver specific options
        "timeout": 60 * 60,
        "namespace": "masonite4",
    },
}



from src.masonite.drivers.session import CookieDriver
from tests import TestCase


class TestSession(TestCase):
    def test_default_sessiondriver(self):
        session = self.application.make("session")
        session.start("cookie")
        driver = session.get_driver()
        self.assertIsInstance(driver, CookieDriver)



from typing import TYPE_CHECKING
from tests import TestCase
from tests.integrations.app.SayHi import SayHello
from tests.integrations.app.GreetingService import GreetingService
from tests.integrations.app.User import User

from src.masonite.exceptions import (
    StrictContainerException,
    MissingContainerBindingNotFound,
)

if TYPE_CHECKING:
    from src.masonite.foundation import Application

from src.masonite.foundation import Application


class SomeObject:

    class_attribute = 1

    def __init__(self, attribute):
        self.attribute = attribute


class SomeStringTypeHintedAppObject:
    def __init__(self, app: "Application"):
        self.app = app


class SomeAppObject:
    def __init__(self, app: Application):
        self.app = app


class TestContainer(TestCase):
    def tearDown(self):
        super().tearDown()
        # unbind container.test after each test
        self.application.objects.pop("container.test", None)

    def test_cannot_bind_module(self):
        from src.masonite import foundation

        with self.assertRaises(StrictContainerException):
            self.application.bind("container.test", foundation)

    def test_can_bind_object(self):
        obj = SomeObject("test")
        self.application.bind("container.test", obj)
        bound_obj = self.application.make("container.test")

        self.assertEqual(obj, bound_obj)
        self.assertEqual(obj.class_attribute, 1)
        self.assertEqual(obj.attribute, "test")

        bound_obj.attribute = "test_updated"

        self.assertEqual(
            self.application.make("container.test").attribute, "test_updated"
        )

    def test_can_override_bound_objects(self):
        obj = SomeObject("test")
        self.application.bind("container.test", obj)
        other_obj = SomeObject("test")
        self.application.bind("container.test", other_obj)
        self.assertEqual(self.application.make("container.test"), other_obj)

    def test_container_has_binding(self):
        self.assertFalse(self.application.has("SomeUnexistingBinding"))
        self.assertFalse(self.application.has(TestContainer))

        self.application.bind("container.test", 1)
        self.assertTrue(self.application.has("container.test"))

    def test_container_cannot_resolve_unbound_objects_which_are_not_classes(self):
        with self.assertRaises(MissingContainerBindingNotFound):
            self.application.make("container.test")

    def test_container_can_resolve_classes_without_binding(self):
        # First pass - not in container
        self.assertFalse(self.application.has(SayHello))
        say_hello = self.application.make(SayHello)
        self.assertIsInstance(say_hello, SayHello)

        # Second pass - we still haven't bound it after the above
        self.assertFalse(self.application.has(SayHello))
        say_hello2 = self.application.make(SayHello)
        self.assertIsInstance(say_hello2, SayHello)

        # Container should have resolved different objects
        self.assertIsNot(say_hello, say_hello2)

    def test_container_can_resolve_nested_class_dependencies_without_binding(self):
        # There's an unbound dependency in __init__()
        service = self.application.make(GreetingService)

        # There's another unbound dependency in handle()
        # And that unbound dependency also has an unbound
        # dependency in its __init()
        result = self.application.resolve(getattr(service, "handle"))

        self.assertIsInstance(result, User)
        self.assertEqual("Jack Sparrow", result.name)

        # All resolved but none bound
        self.assertFalse(self.application.has(GreetingService))
        self.assertFalse(self.application.has(SayHello))

    def test_can_collect_bindings(self):
        bindings = self.application.collect("*.location")
        self.assertGreater(len(bindings), 0)
        self.assertIn("config.location", bindings.keys())

    def test_can_resolve_annotated_parameter(self):
        from src.masonite.response import Response

        def my_method(response: Response):
            return "ok"

        self.application.resolve(my_method)

    def test_can_resolve_class_with_str_type_hinted_parameters(self):
        def my_method(test: SomeStringTypeHintedAppObject):
            return test

        with self.assertRaises(TypeError):
            self.application.resolve(my_method)

    def test_can_resolve_class_with_str_type_hinted_parameters_and_pass_parameter(self):
        def my_method(test: SomeStringTypeHintedAppObject):
            return test

        obj = self.application.resolve(my_method, 1)
        self.assertEquals(obj.app + 1, 2)

    def test_can_resolve_class_with_type_hinted_parameters(self):
        def my_method(test: SomeAppObject):
            return test.app

        self.application.resolve(my_method)



from ..configuration import config


class Cookie:
    def __init__(self, name, value, **options):
        self.options = options

        self.name = name
        self.value = value
        self.http_only = self.__get_option('http_only', True)
        self.secure = self.__get_option('secure', False)
        self.expires = self.__get_option('expires', None)
        self.timezone = self.__get_option('timezone', None)
        self.samesite = self.__get_option('samesite', 'Strict')
        self.path = self.__get_option('path', '/')
        self.encrypt = self.__get_option('encrypt', True)

    def render(self):
        response = f"{self.name}={self.value};"
        if self.http_only:
            response += "HttpOnly;"
        if self.secure:
            response += "Secure;"
        if self.expires:
            if self.timezone:
                response += f"Expires={self.expires} {self.timezone};"
            else:
                response += f"Expires={self.expires};"
        if self.path:
            response += f"Path={self.path};"
        if self.samesite:
            response += f"SameSite={self.samesite};"

        return response

    def __get_option(self, key: str, default: any):
        """
            Get cookie options from config/session.py
            if option key found in options then it return that
            if not found in options then it will fetch from config
            if not found in config then use the default value
        """
        if key in self.options:
            return self.options[key]
        else:
            cookie = config('session.drivers.cookie')
            return cookie[key] if key in cookie else default



from src.masonite.notification import Notification
from src.masonite.mail import Mailable
from src.masonite.notification import Sms

class TestNotification(Notification, Mailable):

    def to_vonage(self, notifiable):
        return Sms().text("Test message" )

    def via(self, notifiable):
        return ["vonage"]


from ..facades import Session


def old(key: str, default: str = "") -> str:
    """Helper used to get value from session for the given key, with a default option."""
    return Session.get(key) or default



import sys
from tests import TestCase

from src.masonite.rates import UnlimitedLimiter, GlobalLimiter, Limit, GuestsOnlyLimiter
from tests.integrations.app.User import User


class TestLimits(TestCase):
    def test_limit_from_str(self):
        limit = Limit.from_str("10/hour")
        assert limit.max_attempts == 10
        assert limit.delay == 60

    def test_limit_per_minute(self):
        limit = Limit.per_minute(5)
        assert limit.max_attempts == 5
        assert limit.delay == 1

    def test_limit_per_hour(self):
        limit = Limit.per_hour(10)
        assert limit.max_attempts == 10
        assert limit.delay == 60

    def test_limit_per_day(self):
        limit = Limit.per_day(10)
        assert limit.max_attempts == 10
        assert limit.delay == 60 * 24

    def test_limit_custom(self):
        limit = Limit(500, 40)
        assert limit.max_attempts == 500
        assert limit.delay == 40

    def test_limit_unlimited(self):
        limit = Limit.unlimited()
        assert limit.max_attempts == sys.maxsize
        assert limit.is_unlimited()


class TestLimiters(TestCase):
    def test_unlimited(self):
        request = "fake"
        limiter = UnlimitedLimiter()
        limit = limiter.allow(request)
        assert limit.is_unlimited()
        assert limit.max_attempts == sys.maxsize

    def test_global(self):
        request = "fake"
        limiter = GlobalLimiter("3/minute")
        limit = limiter.allow(request)
        assert not limit.is_unlimited()
        assert limit.max_attempts == 3
        assert limit.delay == 1
        limiter = GlobalLimiter("100/day")
        limit = limiter.allow(request)
        assert not limit.is_unlimited()
        assert limit.max_attempts == 100
        assert limit.delay == 24 * 60

    def test_guests_only(self):
        # request as guest
        request = self.make_request()
        request._ip = "127.0.0.1"

        limiter = GuestsOnlyLimiter("2/hour")
        limit = limiter.allow(request)
        assert not limit.is_unlimited()
        assert limit.max_attempts == 2
        assert limit.delay == 60
        assert limit.key == "127.0.0.1"

        request._ip = "192.168.0.1"
        limit = limiter.allow(request)
        assert limit.key == "192.168.0.1"

    def test_guests_only_with_authenticated_user(self):
        limiter = GuestsOnlyLimiter("2/hour")
        user = User.find(1)
        # request as authenticated user
        request = self.make_request()
        request.set_user(user)
        limit = limiter.allow(request)
        assert limit.is_unlimited()
